{-# OPTIONS_GHC -ffi #-}
{-# CFILES x87extended_ffi.c #-}
module Data.X87Extended(Extended, extended, extendedToDouble) where

import Data.Typeable
import Foreign.Ptr
import Foreign.ForeignPtr
import Foreign.C
import Foreign.C.String
import Foreign.Storable
import System.IO.Unsafe

newtype X87ExtendedValue = X87ExtendedValue ()
newtype Extended = Extended (ForeignPtr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_alloc" 
	x87extended_alloc :: CDouble -> IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_alloc_pi" 
	x87extended_alloc_pi :: IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_alloc_read" 
	x87extended_alloc_read :: Ptr CChar -> IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_free"
	x87extended_free :: Ptr X87ExtendedValue -> IO ()

foreign import stdcall "x87extended_ffi.h x87extended_to_double" 
	x87extended_to_double :: Ptr X87ExtendedValue -> IO CDouble

foreign import stdcall "x87extended_ffi.h x87extended_show" 
	x87extended_show :: Ptr CChar -> Ptr X87ExtendedValue -> IO CSize

foreign import stdcall "x87extended_ffi.h x87extended_compare" 
	x87extended_compare :: Ptr X87ExtendedValue -> Ptr X87ExtendedValue -> IO CInt

foreign import stdcall "x87extended_ffi.h x87extended_alloc_neg"
	x87extended_alloc_neg :: Ptr X87ExtendedValue -> IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_alloc_abs"
	x87extended_alloc_abs :: Ptr X87ExtendedValue -> IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_alloc_add"
	x87extended_alloc_add :: Ptr X87ExtendedValue -> Ptr X87ExtendedValue -> IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_alloc_sub"
	x87extended_alloc_sub :: Ptr X87ExtendedValue -> Ptr X87ExtendedValue -> IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_alloc_mul"
	x87extended_alloc_mul :: Ptr X87ExtendedValue -> Ptr X87ExtendedValue -> IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_alloc_div"
	x87extended_alloc_div :: Ptr X87ExtendedValue -> Ptr X87ExtendedValue -> IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_alloc_peek"
	x87extended_alloc_peek :: Ptr X87ExtendedValue -> CPtrdiff -> IO (Ptr X87ExtendedValue)

foreign import stdcall "x87extended_ffi.h x87extended_poke"
	x87extended_poke :: Ptr X87ExtendedValue -> CPtrdiff -> Ptr X87ExtendedValue -> IO ()

foreign import stdcall "x87extended_ffi.h x87extended_vorzeichen"
	x87extended_vorzeichen :: Ptr X87ExtendedValue -> IO CInt

foreign import stdcall "x87extended_ffi.h x87extended_exponent"
	x87extended_exponent :: Ptr X87ExtendedValue -> IO CInt

foreign import stdcall "x87extended_ffi.h x87extended_mantisse"
	x87extended_mantisse :: Ptr X87ExtendedValue -> IO CULLong

type Finalize = Ptr X87ExtendedValue -> IO ()
foreign import stdcall "wrapper" 
	toFinalizer :: Finalize -> IO (FunPtr Finalize)

wrapExtended :: Ptr X87ExtendedValue -> IO Extended
wrapExtended v = do
	f <- toFinalizer x87extended_free
	r <- newForeignPtr f v
	return (Extended r)

doubleToExtended :: Double -> Extended
doubleToExtended x = unsafePerformIO $ 
	x87extended_alloc (realToFrac x) >>= wrapExtended

extendedToDouble :: Extended -> Double
extendedToDouble (Extended x) = unsafePerformIO $
	withForeignPtr x $ \y -> x87extended_to_double y >>= return . realToFrac

strToExtended :: [Char] -> Extended
strToExtended s = unsafePerformIO $ 
	(withCString s $ \sd -> x87extended_alloc_read sd) >>= wrapExtended

compareExtended :: Extended -> Extended -> CInt
compareExtended (Extended x) (Extended y) = unsafePerformIO $
	withForeignPtr x $ \xd -> withForeignPtr y $ \yd -> x87extended_compare xd yd

extended = doubleToExtended

instance Show Extended where
	show (Extended x) = unsafePerformIO $ withForeignPtr x $ \xd -> do
		buf <- newCString (take 40 (repeat ' ')) -- ͓K, Extended̍ő傪23Ȃ̂ł܂40΁cc
		len <- x87extended_show buf xd
		peekCStringLen (buf, fromIntegral len)

instance Read Extended where
	readsPrec _ = \s -> let (x, y) = takeExtendedStr s "" "" in [(strToExtended (reverse x), y)]
		where
			takeExtendedStr s x y = case s of
				"" -> (x, y)
				c:cs -> 
					if ((c >= '0') && (c <= '9')) || (c == '.') || (c == 'e') || (c == '+') || (c == '-')
						then takeExtendedStr cs (c:x) y
						else (x, y)

instance Eq Extended where
	(==) (Extended x) (Extended y) = unsafePerformIO $
		withForeignPtr x $ \xd -> withForeignPtr y $ \yd -> x87extended_compare xd yd >>= return . (== 0)

instance Ord Extended where
	compare x y = let r = compareExtended x y in if (r < 0) then LT else if (r > 0) then GT else EQ
	(<) x y = (compareExtended x y) < 0
	(>) x y = (compareExtended x y) > 0
	(<=) x y = (compareExtended x y) <= 0
	(>=) x y = (compareExtended x y) >= 0

instance Num Extended where
	(+) (Extended x) (Extended y) = unsafePerformIO $
		(withForeignPtr x $ \xd -> withForeignPtr y $ \yd -> x87extended_alloc_add xd yd) >>= wrapExtended
	(-) (Extended x) (Extended y) = unsafePerformIO $
		(withForeignPtr x $ \xd -> withForeignPtr y $ \yd -> x87extended_alloc_sub xd yd) >>= wrapExtended
	(*) (Extended x) (Extended y) = unsafePerformIO $
		(withForeignPtr x $ \xd -> withForeignPtr y $ \yd -> x87extended_alloc_mul xd yd) >>= wrapExtended
	negate (Extended x) = unsafePerformIO $
		(withForeignPtr x $ \xd -> x87extended_alloc_neg xd) >>= wrapExtended
	abs (Extended x) = unsafePerformIO $
		(withForeignPtr x $ \xd -> x87extended_alloc_abs xd) >>= wrapExtended
	signum x = fromIntegral (compareExtended x (doubleToExtended 0.0))
	fromInteger x = doubleToExtended (fromInteger x) -- ###double####

instance Real Extended where
	toRational i = toRational (extendedToDouble i) -- ###double####

instance Enum Extended where
	succ i = i + (doubleToExtended 1.0)
	pred i = i - (doubleToExtended 1.0)
	toEnum x = doubleToExtended (toEnum x) -- ###double####
	fromEnum i = fromEnum (extendedToDouble i) -- ###double####

instance Storable Extended where
	sizeOf x = 12 
	alignment x = 4
	peekElemOff a i = x87extended_alloc_peek (castPtr a) (fromIntegral i) >>= wrapExtended
	pokeElemOff a i (Extended x) = withForeignPtr x $ \xd ->
		x87extended_poke (castPtr a) (fromIntegral i) xd

tyConExtended = mkTyCon "Extended"
instance Typeable Extended where
	typeOf _ = mkTyConApp tyConExtended [] 

instance Fractional Extended where
	(/) (Extended x) (Extended y) = unsafePerformIO $
		(withForeignPtr x $ \xd -> withForeignPtr y $ \yd -> x87extended_alloc_div xd yd) >>= wrapExtended
	recip x = doubleToExtended (recip (extendedToDouble x)) -- ###double####
	fromRational r = doubleToExtended (fromRational r) -- ###double####

instance Floating Extended where
	pi = unsafePerformIO $ x87extended_alloc_pi >>= wrapExtended
	exp x = doubleToExtended (exp (extendedToDouble x)) -- ###double####
	log x = doubleToExtended (log (extendedToDouble x)) -- ###double####
	sin x = doubleToExtended (sin (extendedToDouble x)) -- ###double####
	cos x = doubleToExtended (cos (extendedToDouble x)) -- ###double####
	asin x = doubleToExtended (asin (extendedToDouble x)) -- ###double####
	atan x = doubleToExtended (atan (extendedToDouble x)) -- ###double####
	acos x = doubleToExtended (acos (extendedToDouble x)) -- ###double####
	sinh x = doubleToExtended (sinh (extendedToDouble x)) -- ###double####
	cosh x = doubleToExtended (cosh (extendedToDouble x)) -- ###double####
	asinh x = doubleToExtended (asinh (extendedToDouble x)) -- ###double####
	atanh x = doubleToExtended (atanh (extendedToDouble x)) -- ###double####
	acosh x = doubleToExtended (acosh (extendedToDouble x)) -- ###double####

instance RealFrac Extended where
	properFraction x = let (m,y) = properFraction (extendedToDouble x) in (m, doubleToExtended y) -- ###double####
	truncate x = truncate (extendedToDouble x) -- ###double####
	round x = round (extendedToDouble x) -- ###double####
	ceiling x = ceiling (extendedToDouble x) -- ###double####
	floor x = floor (extendedToDouble x) -- ###double####

rareMantisse :: Extended -> Integer
rareMantisse (Extended x) = unsafePerformIO $
	withForeignPtr x $ \xd -> do 
		r <- x87extended_mantisse xd
		s <- x87extended_vorzeichen xd
		return (let sr = fromIntegral r in if s == 0 then sr else -sr)

rareExponent :: Extended -> Int
rareExponent (Extended x) = unsafePerformIO $
	withForeignPtr x $ \xd -> x87extended_exponent xd >>= (return . fromIntegral)

instance RealFloat Extended where
	floatRadix x = floatRadix (extendedToDouble x)  -- ###double####
	floatDigits x = floatDigits (extendedToDouble x) -- ###double####
	floatRange x = floatRange (extendedToDouble x) -- ###double####
	decodeFloat x = (rareMantisse x, rareExponent x)
	encodeFloat m n = doubleToExtended (encodeFloat m n)  -- ###double####
	exponent x = exponent (extendedToDouble x) -- ###double####
	significand x = doubleToExtended (significand (extendedToDouble x)) -- ###double####
	scaleFloat n x = doubleToExtended (scaleFloat n (extendedToDouble x)) -- ###double####
	isNaN x = isNaN (extendedToDouble x) -- ###double####
	isInfinite x = isInfinite (extendedToDouble x) -- ###double####
	isDenormalized x = isDenormalized (extendedToDouble x) -- ###double####
	isNegativeZero x = isNegativeZero (extendedToDouble x) -- ###double####
	isIEEE x = isIEEE (extendedToDouble x) -- ###double####
	atan2 x y = doubleToExtended (atan2 (extendedToDouble x) (extendedToDouble y)) -- ###double####
