#include "x87extended_ffi.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* Runtime */

#ifdef __GNUC__

void __stdcall set8087cw(uint16_t cw)
{
	__asm__("lea 8(%ebp), %eax");
	__asm__("fnclex");
	__asm__("fldcw (%eax)");
}

uint16_t __stdcall get8087cw(void)
{
	__asm__("push %eax");
	__asm__("fnstcw (%esp)");
	__asm__("pop %eax");
}

void __stdcall storet0(long double *x)
{
	__asm__("mov 8(%ebp), %eax");
	__asm__("fldz");
	__asm__("fstpt (%eax)");
}

void __stdcall storet1(long double *x)
{
	__asm__("mov 8(%ebp), %eax");
	__asm__("fld1");
	__asm__("fstpt (%eax)");
}

void __stdcall storetpi(long double *x)
{
	__asm__("mov 8(%ebp), %eax");
	__asm__("fldpi");
	__asm__("fstpt (%eax)");
}

void __stdcall storetd(long double *x, double const *y)
{
	__asm__("mov 12(%ebp), %edx");
	__asm__("mov 8(%ebp), %eax");
	__asm__("fldl (%edx)");
	__asm__("fstpt (%eax)");
}

void __stdcall storedt(double *x, long double const *y)
{
	__asm__("mov 12(%ebp), %edx");
	__asm__("mov 8(%ebp), %eax");
	__asm__("fldt (%edx)");
	__asm__("fstpl (%eax)");
}

int __stdcall comparet(long double const *x, long double const *y)
{
	__asm__("mov 12(%ebp), %edx");
	__asm__("mov 8(%ebp), %eax");
	__asm__("fldt (%edx)");
	__asm__("fldt (%eax)");
	__asm__("fcompp");
	__asm__("fnstsw %ax");
	__asm__("sahf");
	__asm__("seta %al");
	__asm__("setb %dl");
	__asm__("sub %dl, %al");
	__asm__("cbw");
	__asm__("cwde");
}

void __stdcall chst(long double *x, long double const *y)
{
	__asm__("mov 12(%ebp), %edx");
	__asm__("mov 8(%ebp), %eax");
	__asm__("fldt (%edx)");
	__asm__("fchs");
	__asm__("fstpt (%eax)");
}

void __stdcall abst(long double *x, long double const *y)
{
	__asm__("mov 12(%ebp), %edx");
	__asm__("mov 8(%ebp), %eax");
	__asm__("fldt (%edx)");
	__asm__("fabs");
	__asm__("fstpt (%eax)");
}

void __stdcall addt(long double *x, long double const *y, long double const *z)
{
	__asm__("mov 12(%ebp), %eax");
	__asm__("mov 16(%ebp), %edx");
	__asm__("fldt (%eax)");
	__asm__("fldt (%edx)");
	__asm__("mov 8(%ebp), %eax");
	__asm__("faddp");
	__asm__("fstpt (%eax)");
}

void __stdcall subt(long double *x, long double const *y, long double const *z)
{
	__asm__("mov 12(%ebp), %eax");
	__asm__("mov 16(%ebp), %edx");
	__asm__("fldt (%eax)");
	__asm__("fldt (%edx)");
	__asm__("mov 8(%ebp), %eax");
	__asm__("fsubp");
	__asm__("fstpt (%eax)");
}

void __stdcall mult(long double *x, long double const *y, long double const *z)
{
	__asm__("mov 12(%ebp), %eax");
	__asm__("mov 16(%ebp), %edx");
	__asm__("fldt (%eax)");
	__asm__("fldt (%edx)");
	__asm__("mov 8(%ebp), %eax");
	__asm__("fmulp");
	__asm__("fstpt (%eax)");
}

void __stdcall divt(long double *x, long double const *y, long double const *z)
{
	__asm__("mov 12(%ebp), %eax");
	__asm__("mov 16(%ebp), %edx");
	__asm__("fldt (%eax)");
	__asm__("fldt (%edx)");
	__asm__("mov 8(%ebp), %eax");
	__asm__("fdivp");
	__asm__("fstpt (%eax)");
}

#endif

/* FFI */

typedef enum {
	fpu_single = 0,
	fpu_reserved = 1,
	fpu_double = 2,
	fpu_extended = 3,
} fpu_precision_mode ;

static fpu_precision_mode set_precision_mode(fpu_precision_mode mode)
{
	unsigned short cw = get8087cw();
	unsigned short result = (cw >> 8) & 3;
	set8087cw((cw & 0xfcff) | ((unsigned short)mode << 8));
	return result;
}

#define restore_precision_mode set_precision_mode

x87extended * __stdcall x87extended_alloc(double x)
{
	x87extended *result = malloc(sizeof(x87extended));
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	storetd((long double *)result, &x);
	restore_precision_mode(mode);
	return result;
}

x87extended * __stdcall x87extended_alloc_pi(void)
{
	x87extended *result = malloc(sizeof(x87extended));
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	storetpi((long double *)result);
	restore_precision_mode(mode);
	return result;
}

x87extended * __stdcall x87extended_alloc_read(char *s)
{
	x87extended *result = malloc(sizeof(x87extended));
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	double v = strtod(s, NULL);
	storetd((long double *)result, &v);
	restore_precision_mode(mode);
	return result;
}

void __stdcall x87extended_free(x87extended *x)
{
	free(x);
}

double __stdcall x87extended_to_double(x87extended *x)
{
	double result;
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	if(x){
		storedt(&result, (long double *)x);
	}else{
		result = 0.0;
	}
	restore_precision_mode(mode);
	return result;
}

size_t __stdcall x87extended_show(char *s, x87extended *x)
{
	size_t result;
	ptrdiff_t i, j;
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	double v;
	storedt(&v, (long double *)x);
	result = sprintf(s, "%.23e", v); /* Both MS-C and gcc of GHC does not supports Lf... */
	restore_precision_mode(mode);
	for(i = result - 1; i > 0; --i){
		if(s[i] == 'e'){
			for(j = i - 1; j > 0 && s[j] == '0' && s[j - 1] != '.'; --j);
			memmove(&s[j + 1], &s[i], result - i + 1);
			result -= i - j - 1;
			break;
		}
	}
	return result;
}

int __stdcall x87extended_compare(x87extended *x, x87extended *y)
{
	int result;
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	result = comparet((long double *)x, (long double *)y);
	restore_precision_mode(mode);
	return result;
}

x87extended * __stdcall x87extended_alloc_neg(x87extended *x)
{
	x87extended *result = malloc(sizeof(x87extended));
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	chst((long double *)result, (long double *)x);
	restore_precision_mode(mode);
	return result;
}

x87extended * __stdcall x87extended_alloc_abs(x87extended *x)
{
	x87extended *result = malloc(sizeof(x87extended));
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	abst((long double *)result, (long double *)x);
	restore_precision_mode(mode);
	return result;
}

x87extended * __stdcall x87extended_alloc_add(x87extended *x, x87extended *y)
{
	x87extended *result = malloc(sizeof(x87extended));
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	addt((long double *)result, (long double *)x, (long double *)y);
	restore_precision_mode(mode);
	return result;
}

x87extended * __stdcall x87extended_alloc_sub(x87extended *x, x87extended *y)
{
	x87extended *result = malloc(sizeof(x87extended));
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	subt((long double *)result, (long double *)x, (long double *)y);
	restore_precision_mode(mode);
	return result;
}

x87extended * __stdcall x87extended_alloc_mul(x87extended *x, x87extended *y)
{
	x87extended *result = malloc(sizeof(x87extended));
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	mult((long double *)result, (long double *)x, (long double *)y);
	restore_precision_mode(mode);
	return result;
}

x87extended * __stdcall x87extended_alloc_div(x87extended *x, x87extended *y)
{
	x87extended *result = malloc(sizeof(x87extended));
	fpu_precision_mode mode = set_precision_mode(fpu_extended);
	divt((long double *)result, (long double *)x, (long double *)y);
	restore_precision_mode(mode);
	return result;
}

x87extended * __stdcall x87extended_alloc_peek(x87extended *a, ptrdiff_t i)
{
	x87extended *result = malloc(sizeof(x87extended));
	*result = a[i];
	return result;
}

void __stdcall x87extended_poke(x87extended *a, ptrdiff_t i, x87extended *x)
{
	a[i] = *x;
}

int __stdcall x87extended_vorzeichen(x87extended *x)
{
	return ((*(((uint8_t *)x) + 9)) & 0x80) >> 7;
}

int __stdcall x87extended_exponent(x87extended *x)
{
	int16_t result = ((*(int16_t *)(((uint8_t *)x) + 8)) & 0x7fff) - 0x403e;
	return result;
}

unsigned long long __stdcall x87extended_mantisse(x87extended *x)
{
	return *(uint64_t *)x;
}

